# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from resonate.library.exceptions import GenericException

def compress_detections(detections, timefilter=3600):

    '''
    Creates compressed dataframe from detection dataframe

    :param detections: detection dataframe
    :param timefilter: A maximum amount of time in seconds that can pass before
        a new detction event is started
    :return: A Pandas DataFrame matrix of detections events
    '''

    if not isinstance(detections, pd.DataFrame):
        raise GenericException('input parameter must be a Pandas dataframe')

    mandatory_columns = set(['datecollected', 'catalognumber', 'unqdetecid'])

    if mandatory_columns.issubset(detections.columns):

        # Get unique list of animals (not tags), set indices to respect animal and date of detections
        anm_list = detections['catalognumber'].unique()
		# Can't guarantee they've sorted by catalognumber in the input file.
        detections.sort_values(['catalognumber', 'datecollected'], inplace=True)

        # Set up empty data structures and the animal-based groupby object
        detections['seq_num'] = np.nan
        anm_group = detections.groupby('catalognumber')
        out_df = pd.DataFrame()

        # for each animal's detections ordered by time, when station changes, seqnum is incremented.
        for catalognum in anm_list:
            a = anm_group.get_group(catalognum).copy(deep=True)
            # Some say I'm too cautious. Shift logic requires this sort to be true, though.
            a.sort_values(['datecollected', 'station'], inplace=True)
            a.datecollected = pd.to_datetime(a.datecollected)
            a['seq_num'] = ((a.station.shift(1) != a.station) | (a.datecollected.diff().dt.total_seconds() > timefilter)).astype(int).cumsum()
            out_df=out_df.append(a)

        stat_df = out_df.groupby(['catalognumber','seq_num']).agg({  'datecollected':['min', 'max'],
                                                                    'unqdetecid':['first','last'],
                                                                    'seq_num': 'count'})

        # Flatten the multi-index into named columns and cast dates to date objects
        stat_df.columns = ['_'.join(col).strip() for col in stat_df.columns.values]
        stat_df.datecollected_max = pd.to_datetime(stat_df.datecollected_max)
        stat_df.datecollected_min = pd.to_datetime(stat_df.datecollected_min)

        # Calculate average time between detections
		# If it's a single detection, will be 0/1
        stat_df['avg_time_between_det'] = (stat_df['datecollected_max'] - stat_df['datecollected_min']) / np.maximum(1, stat_df['seq_num_count']-1)

        stat_df.rename(columns={'seq_num_count': 'total_count', 'datecollected_max': 'enddate',
								'datecollected_min':'startdate', 'unqdetecid_first': 'startunqdetecid',
								'unqdetecid_last':'endunqdetecid'}, inplace=True)
        # Reduce indexes to regular columns for joining against station number.
        stat_df.reset_index(inplace=True)

        # Join stations to result. Could add lat/lon here as well.
        out_df = out_df[['catalognumber', 'station', 'seq_num']].drop_duplicates().merge(stat_df, on=['catalognumber', 'seq_num'])

        return out_df
    else:
        raise GenericException("Missing required input columns: {}".format(mandatory_columns - set(detections.columns)))



__author__ = "Alex Nunes"
__credits__ = ["Alex Nunes", "Jon Pye", "Brian Jones", "Marta Mihoff"]
__license__ = "GPL"
__version__ = "0.3.2"
__maintainer__ = "Alex Nunes"
__email__ = "anunes@dal.ca"
__status__ = "Production"




import pandas as pd


def add_unqdetecid(input_file, encoding='utf-8-sig'):
    """
    Adds the unqdetecid column to an input csv file. The resulting file is returned as a pandas DataFrame object.

    :param input_file: Path to the input csv file.
    :param encoding: source encoding for the input file (Default utf8-bom)
    :return: padnas DataFrame including unqdetecid column.
    """
    if isinstance(input_file, pd.DataFrame):
        input_df = input_file
        input_df['unqdetecid'] = input_df.index+1
    else:
        input_df = pd.read_csv(input_file, encoding=encoding)
        input_df['unqdetecid'] = input_df.index+1
    return input_df



from resonate.filter_detections import get_distance_matrix
from datetime import datetime, timedelta
import pandas as pd

def velocity_filter(detections, max_vel=10):
    pd.options.mode.chained_assignment = None
    dm = get_distance_matrix(detections)

    lead_lag_df = pd.DataFrame()

    for index, group in detections.sort_values(['datecollected']).groupby(['catalognumber']):
        group['lag_station'] = group.station.shift(1).fillna(group.station)
        group['lead_station'] = group.station.shift(-1).fillna(group.station)

        group.datecollected = pd.to_datetime(group.datecollected)
        group['lag_time_diff'] = group.datecollected.diff().fillna(timedelta(seconds=1))
        group['lead_time_diff'] = group.lag_time_diff.shift(-1).fillna(timedelta(seconds=1))

        lead_lag_df = lead_lag_df.append(group)

    del detections

    vel_df = pd.DataFrame()
    for index, group in lead_lag_df.groupby(['station', 'lag_station', 'lead_station']):
        stn = group.station.unique()[0]
        lag_stn = group.lag_station.unique()[0]
        lead_stn = group.lead_station.unique()[0]

        lag_distance = dm.loc[stn, lag_stn].m
        lead_distance = dm.loc[stn, lead_stn].m

        group['lag_distance_m'] = lag_distance
        group['lead_distance_m'] = lead_distance
        vel_df = vel_df.append(group)

    del lead_lag_df

    vel_df['lag_velocity'] = vel_df.lag_distance_m/vel_df.lag_time_diff.dt.total_seconds()
    vel_df['lead_velocity'] = vel_df.lead_distance_m/vel_df.lead_time_diff.dt.total_seconds()
    vel_df.sort_index(inplace=True)
    filtered_detections = dict()
    filtered_detections['filtered'] = vel_df[(vel_df.lag_velocity <= max_vel) & (vel_df.lead_velocity <= max_vel)].reset_index(drop=True)
    filtered_detections['suspect'] = vel_df[(vel_df.lag_velocity > max_vel) | (vel_df.lead_velocity > max_vel)].reset_index(drop=True)
    return filtered_detections



import pandas as pd
import geopy

def interval_data(compressed_df, dist_matrix_df, station_radius_df=None):
    """
    Creates a detection interval file from a compressed detection, distance matrix and station detection radius DataFrames

    :param compressed_df: compressed detection dataframe
    :param dist_matrix_df: station distance matrix dataframe
    :param station_radius: station distance radius dataframe
    :return: interval detection Dataframe
    """

    # Create two dataframes from input compressed detections and decrement the second's seq_num
    fst = compressed_df[
        ['catalognumber', 'station', 'seq_num', 'total_count', 'startdate', 'enddate', 'endunqdetecid']].copy()
    snd = compressed_df[
        ['catalognumber', 'station', 'seq_num', 'total_count', 'startdate', 'enddate', 'endunqdetecid']].copy()
    snd.seq_num -= 1

    # Rename columns
    fst.columns = ['catalognumber', 'from_station', 'seq_num', 'from_detcnt',
                   'from_arrive', 'from_leave', 'unqdetid_from']
    snd.columns = ['catalognumber', 'to_station', 'seq_num', 'to_detcnt',
                   'to_arrive', 'to_leave', 'unqdetid_arrive']

    # Merge the two DataFrames together linking catalognumber and seq_num
    merged = pd.merge(fst, snd, how='left', on=['catalognumber', 'seq_num'])

    # Create additional column placeholders
    merged['intervaltime'] = None
    merged['intervalseconds'] = None
    merged['distance_m'] = None
    merged['metres_per_second'] = None

    # Loop through all rows linking distance matrices and calculating average speed between intervals
    for idx, item in merged.iterrows():
        # If any of the station pairs don't exist, skip processing current row
        if not (pd.isnull(item['from_station']) or pd.isnull(item['to_station'])):
            # Get station matrix distance (input in km)
            matrix_distance_km = dist_matrix_df.loc[item['from_station'], item['to_station']]

            # If matrix pair exists do distance calculations
            if matrix_distance_km:
                if isinstance(station_radius_df, pd.DataFrame):
                    stn1_radius = station_radius_df.loc[item['from_station'], 'radius']
                    stn2_radius = station_radius_df.loc[item['to_station'], 'radius']

                    distance = max(geopy.distance.Distance(0).km, matrix_distance_km - stn1_radius.km - stn2_radius.km)*1000
                else:
                    distance = max(geopy.distance.Distance(0).km, matrix_distance_km)*1000

                merged.loc[idx, 'distance_m'] =  distance

                time_interval = item['to_arrive'] - item['from_leave']
                merged.loc[idx, 'intervaltime'] =  time_interval
                merged.loc[idx, 'intervalseconds'] = time_interval.total_seconds()

                if time_interval.seconds != 0:
                    merged.loc[idx, 'metres_per_second']= distance / time_interval.seconds

    return merged



import pandas as pd
import plotly.offline as py
import plotly.graph_objs as go
from resonate.library.exceptions import GenericException

def abacus_plot(detections, ycolumn='catalognumber', color_column=None,ipython_display=True, title = 'Abacus Plot', filename=None):

    '''
    Creates a plotly abacus plot from a pandas dataframe

    :param detections: detection dataframe
    :param ycolumn: the series/column for the y axis of the plot
    :param color_column: the series/column to group by and assign a color
    :param ipython_display: a boolean to show in a notebook
    :param title: the title of the plot
    :param filename: Plotly filename to write to

    :return: A plotly scatter plot
    '''

    if not isinstance(detections, pd.DataFrame):
        raise GenericException('input parameter must be a Pandas dataframe')

    mandatory_columns = set(['datecollected', ycolumn])

    if color_column is not None:
        mandatory_columns.add(color_column)

    if mandatory_columns.issubset(detections.columns):

        detections = detections[~detections.unqdetecid.str.contains('release')].reset_index(drop=True)

        if color_column is not None:
            data = list()

            for group in detections.groupby(color_column):
                data.append(
                    {
                        'x': group[1].datecollected.tolist(),
                        'y': group[1][ycolumn].tolist(),
                        'mode': 'markers',
                        'name': group[0]
                    }
                )
        else:
            data = [
                {
                    'x': detections.datecollected.tolist(),
                    'y': detections[ycolumn].tolist(),
                    'mode': 'markers',
                }
            ]

        layout = dict(
            title = title,
            xaxis=dict(
                autorange=False,
                range=[detections.datecollected.min(), detections.datecollected.max()]
            ),
            yaxis=dict(
                autorange=True
            ),
            margin= dict(
                l=175
            )
        )

        fig = { 'data':data, 'layout':layout }

        if ipython_display:
            py.init_notebook_mode()
            return py.iplot(fig)
        else:
            return py.plot(fig, filename=filename)
    else:
        raise GenericException("Missing required input columns: {}".format(mandatory_columns - set(detections.columns)))



import pandas as pd
import plotly.offline as py
import plotly.graph_objs as go
from resonate.library.exceptions import GenericException

def bubble_plot(detections, type='detections', ipython_display=True,
                title = 'Bubble Plot', height=700, width=1000,
                plotly_geo=None, filename=None, mapbox_token=None,
                marker_size=10,colorscale='Viridis'):

    '''
    Creates a plotly abacus plot from a pandas dataframe

    :param detections: detection dataframe
    :param ipython_display: a boolean to show in a notebook
    :param title: the title of the plot
    :param height: the height of the plotl
    :param width: the width of the plotly
    :param plotly_geo: an optional dictionary to controle the
        geographix aspects of the plot
    :param filename: Plotly filename to write to
    :param mapbox_token: A string of mapbox access token
    :param marker_size: An int to indicate the diameter in pixels
    :param colorscale: A string to indicate the color index

    :return: A plotly geoscatter plot or mapbox plot
    '''

    if not isinstance(detections, pd.DataFrame):
        raise GenericException('input parameter must be a Pandas dataframe')

    mandatory_columns = set(['station', 'catalognumber', 'unqdetecid', 'latitude', 'longitude', 'datecollected'])


    if mandatory_columns.issubset(detections.columns):
        detections = detections[['station', 'catalognumber', 'unqdetecid', 'latitude', 'longitude', 'datecollected']].reset_index(drop=True)


        if type =='individual':
            detections = detections.drop(['unqdetecid', 'datecollected'],axis=1).drop_duplicates()
        detections = detections.groupby(['station', 'latitude', 'longitude']).size().reset_index(name='counts')

        map_type = 'scattergeo'

        if mapbox_token is not None:
            map_type = 'scattermapbox'
            mapbox=dict(
                accesstoken=mapbox_token,
                center=dict(
                    lon = detections.longitude.mean(),
                    lat = detections.latitude.mean()
                ),
                zoom=5,
                style='light'
            )

        data = [
            {
                'lon': detections.longitude.tolist(),
                'lat': detections.latitude.tolist(),
                'text': detections.station+" : "+detections.counts.astype(str),
                'mode': 'markers',
                'marker': {
                    'color': detections.counts.tolist(),
                    'size':marker_size,
                    'showscale': True,
                    'colorscale':colorscale,
                    'colorbar':{
                        'title':'Detection Count'
                    }
                },
                'type':map_type
            }
        ]

        if plotly_geo is None:
            plotly_geo = dict(
                showland = True,
                landcolor = "rgb(255, 255, 255)",
                showocean = True,
                oceancolor = "rgb(212,212,212)",
                showlakes = True,
                lakecolor = "rgb(212,212,212)",
                showrivers = True,
                rivercolor = "rgb(212,212,212)",
                resolution = 50,
                showcoastlines = False,
                showframe=False,
                projection = dict(
                    type = 'mercator',
                )
            )

        plotly_geo.update(
            center = dict(
                lon = detections.longitude.mean(),
                lat = detections.latitude.mean()
            ),
            lonaxis = dict(
                range= [ detections.longitude.min(), detections.longitude.max()],
            ),
            lataxis = dict (
                range= [ detections.latitude.min(), detections.latitude.max()],
            )
        )


        if mapbox_token is None:
            layout = dict(
                geo = plotly_geo,
                title = title
            )
        else:
            layout = dict(title=title,
                            autosize=True,
                            hovermode='closest',
                            mapbox=mapbox
                        )

        if ipython_display:
            layout.update(
                height=height,
                width=width
            )
            fig = { 'data':data, 'layout':layout }

            py.init_notebook_mode()
            return py.iplot(fig)
        else:
            fig = { 'data':data, 'layout':layout }
            return py.plot(fig, filename=filename)
    else:
        raise GenericException("Missing required input columns: {}".format(mandatory_columns - set(detections.columns)))



import pandas as pd
from datetime import datetime
import resonate.compress as cp
import plotly.offline as py
import plotly.graph_objs as go
import math


def total_days_diff(detections):
    """
    total_days_diff
    Determines the total days difference.

    The difference is determined
    by the minimal startdate of every detection and the maximum enddate of
    every detection. Both are converted into a datetime then subtracted to
    get a timedelta. The timedelta is converted to seconds and divided by
    the number of seconds in a day (86400). The function returns a floating
    point number of days (i.e. 503.76834).

    :param detections: Pandas DataFrame pulled from the compressed detections CSV

    :return: An float in the number of days

    """
    first = datetime.strptime(detections.startdate.min(), "%Y-%m-%d %H:%M:%S")
    last = datetime.strptime(detections.enddate.max(), "%Y-%m-%d %H:%M:%S")
    total = last - first
    total = total.total_seconds()/86400.0
    return total





def total_days_count(detections):
    '''
    total_days_count
    The function below takes a Pandas DataFrame and determines the number of days any
    detections were seen on the array.

    The function converst both the startdate and enddate columns into a date with no hours, minutes,
    or seconds. Next it creates a list of the unique days where a detection was seen. The size of the
    list is returned as the total number of days as an integer.

    *** NOTE ****
    Possible rounding error may occur as a detection on 2016-01-01 23:59:59 and a detection on
    2016-01-02 00:00:01 would be counted as days when it is really 2-3 seconds.


    :param detections: Pandas DataFrame pulled from the compressed detections CSV

    :return: An int in the number of days

    '''
    detections['startdate'] = detections['startdate'].apply(datetime.strptime, args=("%Y-%m-%d %H:%M:%S",)).apply(datetime.date)
    detections['enddate'] = detections['enddate'].apply(datetime.strptime, args=("%Y-%m-%d %H:%M:%S",)).apply(datetime.date)
    detections = pd.unique(detections[['startdate', 'enddate']].values.ravel())
    return detections.size





def aggregate_total_with_overlap(detections):
    '''
    aggregate_total_with_overlap

    The function below aggregates timedelta of startdate and enddate of each detection into
    a final timedelta then returns a float of the number of days. If the startdate and enddate
    are the same, a timedelta of one second is assumed.

    :param detections: Pandas DataFrame pulled from the compressed detections CSV

    :return: An float in the number of days

    '''
    total = pd.Timedelta(0)
    detections['startdate'] = detections['startdate'].apply(datetime.strptime, args=("%Y-%m-%d %H:%M:%S",))
    detections['enddate'] = detections['enddate'].apply(datetime.strptime, args=("%Y-%m-%d %H:%M:%S",))
    detections['timedelta'] = detections['enddate'] - detections['startdate']

    for index, row in detections.iterrows():
        if row['timedelta'] > pd.Timedelta(0):
            diff = row['timedelta']
        else:
            diff = pd.Timedelta('1 second')
        total += diff
    return total.total_seconds()/86400.0





def aggregate_total_no_overlap(detections):
    '''
    aggregate_total_no_overlap

    The function below aggregates timedelta of startdate and enddate, excluding overlap between
    detections. Any overlap between two detections is converted to a new detection using the earlier
    startdate and the latest enddate. If the startdate and enddate are the same, a timedelta of one
    second is assumed.

    :param detections: pandas DataFrame pulled from the compressed detections CSV

    :return: An float in the number of days

    '''
    total = pd.Timedelta(0)

    # sort and convert datetimes
    detections = detections.sort_values(by='startdate', ascending=False).reset_index(drop=True)
    detections['startdate'] = detections['startdate'].apply(datetime.strptime, args=("%Y-%m-%d %H:%M:%S",))
    detections['enddate'] = detections['enddate'].apply(datetime.strptime, args=("%Y-%m-%d %H:%M:%S",))

    # A stack is used as an easy way to organize and maintain the detections
    detection_stack = list(detections.T.to_dict().values())

    # Run the loop while the stack is not empty
    while len(detection_stack) > 0:

        current_time_block = detection_stack.pop()

        # Make sure the current element is not empty
        if current_time_block:

            # Pop the next item if the stack is not empty
            if len(detection_stack) > 0:
                next_time_block = detection_stack.pop()
            else:
                next_time_block = False

            # Check to see if we are down to the last item in the stack or there is no overlap
            if not next_time_block or next_time_block['startdate'] > current_time_block['enddate']:

                # Create the timedelta and add it to the total, assuming 1 second if the timedelta equals 0
                diff = pd.Timedelta(0)
                diff += current_time_block['enddate'] - current_time_block['startdate']
                if diff == pd.Timedelta(0):
                    diff = pd.Timedelta('1 second')
                total += diff

                # Add the next block back into the stack so that it can be used in the next iteration
                detection_stack.append(next_time_block)
            else:

                # If there is overlap take a new endate, eliminating the overlap, and add it back into the stack for the next iteration
                current_time_block['enddate'] = max([current_time_block['enddate'], next_time_block['enddate']])
                detection_stack.append(current_time_block)

    # Return the value as a float in days
    return total.total_seconds()/86400.0





def get_days(dets, calculation_method='kessel'):
    '''
    get_days

    Determines which calculation method to use for the residency index.

    Wrapper method for the calulation methods above.

    :param dets: A Pandas DataFrame pulled from the compressed detections CSV

    :param calculation_method: determines which method above will be used to
        count total time and station time

    :return: An int in the number of days

    '''
    days = 0

    if calculation_method == 'aggregate_with_overlap':
        days = aggregate_total_with_overlap(dets)
    elif calculation_method == 'aggregate_no_overlap':
        days = aggregate_total_no_overlap(dets)
    elif calculation_method == 'timedelta':
        days = total_days_diff(dets)
    else:
        days = total_days_count(dets)

    return days


def get_station_location(station, detections):
    '''
    get_station_location

    Returns the longitude and latitude of a station/receiver given the station
    and the table name.

    :param station: String that contains the station name
    :param detections: the table name in which to find the station

    :return: A Pandas DataFrame of station, latitude, and longitude

    '''
    location = detections[detections.station == station][:1]
    location = location[['station', 'longitude', 'latitude']]
    return location

def plot_ri(ri_data, ipython_display=True,
                    title = 'Bubble Plot', height=700,
                    width=1000, plotly_geo=None, filename=None,
                    marker_size = 6, scale_markers=False,
                    colorscale='Viridis', mapbox_token=None):
    '''
    plot_ri

    :param ri_data: A Pandas DataFrame generated from ``residency_index()``
    :param ipython_display: a boolean to show in a notebook
    :param title: the title of the plot
    :param height: the height of the plotl
    :param width: the width of the plotly
    :param plotly_geo: an optional dictionary to controle the
        geographix aspects of the plot
    :param filename: Plotly filename to write to
    :param mapbox_token: A string of mapbox access token
    :param marker_size: An int to indicate the diameter in pixels
    :param scale_markers: A boolean to indicate whether or not markers are scaled by their value
    :param colorscale: A string to indicate the color index

    :return: A plotly geoscatter

    '''
    ri_data = ri_data.sort_values('residency_index')

    map_type = 'scattergeo'

    if mapbox_token is not None:
        map_type = 'scattermapbox'
        mapbox=dict(
            accesstoken=mapbox_token,
            center=dict(
                lon = ri_data.longitude.mean(),
                lat = ri_data.latitude.mean()
            ),
            zoom=5,
            style='light'
        )

    if scale_markers:
        marker_size = (ri_data.residency_index * marker_size + 5).tolist()
    else:
        marker_size+=5
    data = [
        {
            'lon': ri_data.longitude.tolist(),
            'lat': ri_data.latitude.tolist(),
            'text': ri_data.station+" : "+ri_data.residency_index.astype(str),
            'mode': 'markers',
            'marker': {
                'color': ri_data.residency_index.tolist(),
                'size':marker_size,
                'showscale': True,
                'colorscale':colorscale,
                'colorbar':{
                    'title':'Detection Count'
                }
            },
            'type':map_type
        }
    ]

    if plotly_geo is None:
        plotly_geo = dict(
            showland = True,
            landcolor = "rgb(255, 255, 255)",
            showocean = True,
            oceancolor = "rgb(212,212,212)",
            showlakes = True,
            lakecolor = "rgb(212,212,212)",
            showrivers = True,
            rivercolor = "rgb(212,212,212)",
            resolution = 50,
            showcoastlines = False,
            showframe=False,
            projection = dict(
                type = 'mercator',
            )
        )
    plotly_geo.update(
        center = dict(
            lon = ri_data.longitude.mean(),
            lat = ri_data.latitude.mean()
        ),
        lonaxis = dict(
            range= [ ri_data.longitude.min(), ri_data.longitude.max()],
        ),
        lataxis = dict (
            range= [ ri_data.latitude.min(), ri_data.latitude.max()],
        )
    )


    if mapbox_token is None:
        layout = dict(
            geo = plotly_geo,
            title = title
        )
    else:
        layout = dict(title=title,
                        autosize=True,
                        hovermode='closest',
                        mapbox=mapbox
                    )

    if ipython_display:
        layout.update(
            height=height,
            width=width
        )
        fig = { 'data':data, 'layout':layout }

        py.init_notebook_mode()
        return py.iplot(fig)
    else:
        fig = { 'data':data, 'layout':layout }
        return py.plot(fig, filename=filename)


def residency_index(detections, calculation_method='kessel'):

    '''
    residency_index

    This function takes in a detections CSV and determines the residency
    index for reach station.

    Residence Index (RI) was calculated as the number of days an individual
    fish was detected at each receiver station divided by the total number of
    days the fish was detected anywhere on the acoustic array. - Kessel et al.

    :param detections: CSV Path
    :param calculation_method: determines which method above will be used to
        count total time and station time

    :return: A residence index DataFrame with the following columns

        * days_detected
        * latitude
        * longitude
        * residency_index
        * station

    '''
    dets = cp.compress_detections(detections)

    # Converting start and end date to strings
    dets['startdate'] = dets['startdate'].astype(str)
    dets['enddate'] = dets['enddate'].astype(str)

    # Remove any release locations
    dets = dets[~dets['startunqdetecid'].astype(str).str.contains("release")]

    print('Creating the residency index using the {0} method.\nPlease be patient, I am currently working...'.format(calculation_method))

    # Determine the total days from a copy of the DataFrame
    total_days = get_days(dets.copy(), calculation_method)

    # Init the stations list
    station_list = []

    # For each unique station determine the total number of days there were
    # detections at the station
    for station in dets['station'].unique():
        st_dets = pd.DataFrame(dets[dets['station'] == station])
        total = get_days(st_dets.copy(), calculation_method)
        location = get_station_location(station, detections)
        # Determine the RI and add the station to the list
        station_dict = {'station': station, 'days_detected': total, 'residency_index': (total/(float(total_days))),
                        'longitude': location['longitude'].values[0], 'latitude': location['latitude'].values[0]}
        station_list.append(station_dict)

    # convert the station list to a Dataframe
    all_stations = pd.DataFrame(station_list)

    # sort and reset the index for the station DataFrame
    all_stations = all_stations.sort_values(by='days_detected', ascending=False).reset_index(drop=True)

    print("OK!")
    # Return the stations RI DataFrame
    return all_stations



"""Generic Exception Class."""


class GenericException(Exception):
    """Exceptions."""

    def __init__(self, msg):
        """Init."""
        self.msg = msg

    def __str__(self):
        """STR."""
        return self.msg



from resonate.filter_detections import get_distance_matrix
from datetime import datetime, timedelta
import pandas as pd

def distance_filter(detections, max_dist=100000):
    pd.options.mode.chained_assignment = None
    dm = get_distance_matrix(detections)

    lead_lag_stn_df = pd.DataFrame()

    for index, group in detections.sort_values(['datecollected']).groupby(['catalognumber']):
        group['lag_station'] = group.station.shift(1).fillna(group.station)
        group['lead_station'] = group.station.shift(-1).fillna(group.station)
        lead_lag_stn_df = lead_lag_stn_df.append(group)

    del detections

    distance_df = pd.DataFrame()

    for index, group in lead_lag_stn_df.groupby(['station', 'lag_station', 'lead_station']):
        stn = group.station.unique()[0]
        lag_stn = group.lag_station.unique()[0]
        lead_stn = group.lead_station.unique()[0]
        lag_distance = dm.loc[stn, lag_stn].m
        lead_distance = dm.loc[stn, lead_stn].m
        group['lag_distance_km'] = lag_distance
        group['lead_distance_km'] = lead_distance
        distance_df = distance_df.append(group)

    del lead_lag_stn_df
    distance_df.sort_index(inplace=True)

    filtered_detections = dict()
    filtered_detections['filtered'] = distance_df[(distance_df.lag_distance_km <= max_dist) & (distance_df.lead_distance_km <= max_dist)].reset_index(drop=True)
    filtered_detections['suspect'] = distance_df[(distance_df.lag_distance_km > max_dist) | (distance_df.lead_distance_km > max_dist)].reset_index(drop=True)
    return filtered_detections



import pandas as pd
import os, sys
import datetime
import simplejson as json
import resonate.compress as cp


def unix_time_millis(dt):
    """
    Returns a datetime in milliseconds

    :param dt: datetime/timestamp
    :return: datetime in milliseconds
    """
    epoch = datetime.datetime.utcfromtimestamp(0)
    return (dt - epoch).total_seconds() * 1000.0



def create_geojson(detections, title, dets_table='', inc=5000):
    """
    This function maps a compressed file and converts the necessary fields
    into a GeoJSON file that can be easily read by Leaflet

    :param detections: a compressed or uncompressed csv detections file
    :param dets_table: An override variable if the table does not match the
        file name
    :param inc: the number of detections to include in each subection of
        the json

    :return: JSON object, the filename, center_x, center_y

    """
    dets = cp.compress_detections(detections)

    # Remove any release locations
    dets = dets[~dets['startunqdetecid'].astype(str).str.contains("release")]


    # Get a list of the unique stations
    locs = detections[['station', 'longitude', 'latitude']].drop_duplicates(subset='station')

    # Add the station location to the compressed detections
    data = pd.merge(locs, dets, on='station', how='inner')

    # Convert startdate and enddate to a milliseconds
    data['startdate'] = data['startdate'].map(unix_time_millis)
    data['enddate'] = data['enddate'].map(unix_time_millis)

    # Sort by start date and reset the index to 1
    data = data.sort_values(by='startdate', ascending=True)
    data.reset_index(drop=True, inplace=True)
    data.index += 1

    # Create a hue index for each individual catalognumber and return a dictionary
    hue_increment = 360/data.catalognumber.unique().size
    animals = pd.DataFrame(data.catalognumber.unique())
    animals['hue'] = animals.index*hue_increment
    animals.columns = ['animals','hue']
    animals = animals.set_index(['animals'])
    animals = animals.T.to_dict()

    detection_geojson = []

    # Get center points fro the map
    center_y = data.latitude.median()
    center_x = data.longitude.median()


    start = 1
    end = inc
    cap = 100000

    # Loop through the detections and create sets of geojson detections the size of the increment
    while start < data.index.size and start <= cap:
        geojson ={
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Point",
                        "coordinates": [d["longitude"], d["latitude"]],
                    },
                    "starttime": d['startdate'],
                    "endtime": d['enddate'],
                    "catalognumber": d['catalognumber'],
                    "station": d['station'],
                    'hue': str(animals[d['catalognumber']]['hue'])
                } for index, d in data[start:end].iterrows()]
        }
        # Add the set of detections to the list
        detection_geojson.append(geojson)
        start = end+1
        end += inc

    # Print message if there are more than cap for detections, defaulting to 100000
    if start > cap:
        print("Only first "+str(cap)+" detections used, please subset your data to see more.")

    # Write the geojson out to a json file
    json_name = title.lower().replace(' ', '_')
    print("Writing JSON file to " +json_name+".json")
    output = open("./html/"+json_name+".json", 'w')
    json.dump(detection_geojson, output)
    output.close()

    # Create string with just the file name and no path
    filename = json_name.replace("./html/", '')+".json"

    # Return the json object, filename, and the center points
    return {'json': detection_geojson, 'filename': filename, 'center_x': center_x, 'center_y': center_y}



import pandas as pd


def cohort(compressed_df, interval_time=60):
    """
    Creates a dataframe of cohorts using a compressed detection file

    :param compressed_df: compressed dataframe
    :param interval_time: cohort detection time interval (in minutes)
    :return: cohort dataframe with the following columns

        * anml_1
        * anml_1_seq
        * station
        * anml_2
        * anml_2_seq
        * anml_2_arrive
        * anml_2_depart
        * anml_2_startunqdetecid
        * anml_2_endunqdetecid
        * anml_2_detcount

    """

    # Convert input int interval_time into a timedelta object
    interval_time = pd.to_timedelta(interval_time, unit='m')

    # Sort input compressed data file
    cmps = compressed_df.sort_values(['catalognumber', 'seq_num'])

    # Loop through rows in the compressed data file and choose
    # cohorts if the times are within range
    seen = []
    final_set = []

    for idx1, item1 in cmps.iterrows():
        match = cmps[
            (cmps.station == item1.station) &
            (cmps.catalognumber != item1.catalognumber) &
            (~cmps.index.to_series().apply(lambda x: (x, idx1)).isin(seen)) &
            ((cmps.startdate > item1.startdate - interval_time) &
                (cmps.startdate < item1.enddate + interval_time) |
                (cmps.enddate > item1.startdate - interval_time) &
                (cmps.enddate < item1.enddate + interval_time))]

        if not match.empty:
            match.insert(0, 'anml_1', item1.catalognumber)
            match.insert(1, 'anml_1_seq', item1.seq_num)
            final_set.extend(match[['anml_1', 'anml_1_seq', 'station',
                                    'catalognumber', 'seq_num', 'startdate',
                                    'enddate', 'startunqdetecid',
                                    'endunqdetecid', 'total_count'
                                    ]].values.tolist())
            seen.extend([(idx1, i) for i in match.index.tolist()])

    output_df = pd.DataFrame(final_set)
    output_df.columns = ['anml_1',
                         'anml_1_seq',
                         'station',
                         'anml_2',
                         'anml_2_seq',
                         'anml_2_arrive',
                         'anml_2_depart',
                         'anml_2_startunqdetecid',
                         'anml_2_endunqdetecid',
                         'anml_2_detcount']
    return output_df



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from codecs import open

with open('README', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE', encoding='utf-8') as f:
    license = f.read()

setup(
    name='resonATe',
    version='0.3.2',
    description='resonate data analysis package',
    long_description=readme,
    author='Alex Nunes',
    include_package_data=True,
    author_email='anunes@dal.ca',
    url='https://gitlab.oceantrack.org/otndc/resonate',
    download_url = 'https://gitlab.oceantrack.org/otndc/resonate',
    license=license,
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=('tests', 'docs'))
)



import jinja2 as jin
import pandas as pd
import resonate.geojson as gj
from IPython.display import IFrame
import os

template_file = os.path.join(os.path.dirname(__file__),
                             'templates/leaflet_timeline.html')


def create_leaflet_timeline(title, json, center_y=0, center_x=0, zoom=8,
                            steps=100000, basemap='dark_layer'):
    '''
    Uses Jinja to write an html file that can then be viewed through a browser
    or used with other functions.


    :param title: Pandas DataFrame pulled from the compressed detections CSV
    :param json: JSON file name to use in for the map
    :param center_y: Latitude center for map
    :param center_x: Longitude center for map
    :param steps: The number of increments the slider will snap to
    '''

    template = jin.Template(open(template_file, 'r').read())
    html = template.render(title=title, json_file=json, zoom=zoom, center_y=center_y, center_x=center_x, steps=steps, layer=basemap)
    output = open("./html/"+title+".html", 'w')

    print("Writing html file to ./html/"+title+".html...")

    output.write(html)
    output.close()

    return "HTML file written to ./html/"+title+".html"


def render_map(dets, title, width=900, height=450,
               zoom=8, basemap='dark_layer'):
    """
    Return an IFrame with Leaflet Timeline map of a limited number of
    compressed detections.

    :param det_file: The CSV file of compressed or non-compressed detections to
        be used
    :param title: The title of the html file
    :param width: The width of the iframe
    :param height: The height of the iframe
    :param zoom: The initial zoom of the map

    :return: An iFrame containing the detections map
    """
    # Create html subfolder for output if there's not one already.
    if not os.path.exists('./html'):
        os.makedirs('./html')

    # Create the GeoJSON to be used
    json = gj.create_geojson(dets, title=title)

    if not json:
        print('Unable to create map, please resolve issues listed above.')
        return None

    # Create the HTML and javascript that will be used for the map
    create_leaflet_timeline(json=json['filename'],
                            title=title,
                            center_y=json['center_y'],
                            center_x=json['center_x'],
                            zoom=zoom,
                            basemap=basemap)

    # Create and return the IFrame to be rendered for the user
    iframe = IFrame('./html/'+title+'.html', width=width, height=height)

    return iframe



import pandas as pd
import datetime
import numpy as np
from geopy.distance import vincenty
from resonate.library.exceptions import GenericException


def get_distance_matrix(detectiondf):

    """
    Creates a distance matrix of all stations in the array or line.

    :param detectiondf: a Pandas DataFrame of detections

    :return: A Pandas DataFrame matrix of station to station distances

    """
    stn_grouped = detectiondf.groupby('station')
    stn_locs = stn_grouped[['longitude', 'latitude']].mean()

    dist_mtx = pd.DataFrame(
        np.zeros(len(stn_locs) ** 2).reshape(len(stn_locs), len(stn_locs)),
        index=stn_locs.index, columns=stn_locs.index)

    for cstation in dist_mtx.columns:
        for rstation in dist_mtx.index:
            cpoint = (stn_locs.loc[cstation,'latitude'], stn_locs.loc[cstation,'longitude'])
            rpoint = (stn_locs.loc[rstation,'latitude'], stn_locs.loc[rstation,'longitude'])
            dist_mtx.loc[rstation,cstation]= vincenty(cpoint,rpoint).km
    dist_mtx.index.name = None
    return dist_mtx


def filter_detections(detections, suspect_file=None,
                      detection_radius=None, min_time_buffer=60,
                      distance_matrix=False):

    """
    Filters isolated detections that are more than min_time_buffer apart from
    other dets. for a series of detections in detection_file. Returns Filtered
    and Suspect dataframes.
    suspect_file can be a file of existing suspect detections to remove before
    filtering.
    dist_matrix is created as a matrix of between-station distances from
    stations defined in the input file.

    :param detection_file: A CSV file of acoustic detections

    :param suspect_file: Path to a user specified suspect file

    :param detection_radius:

    :param min_time_buffer: The minimum of time required for outlier detections in
        minutes

    :param distance_matrix: A boolean of whether or not to generate the
        distance matrix

    :return: A list of Pandas DataFrames of filtered detections, suspect
        detections, and a distance matrix

    """

    # Set of mandatory column names for detection_file
    mandatory_columns = set(['station',
                             'unqdetecid',
                             'datecollected',
                             'catalognumber'])

    # Load the file into a dataframe
    df = detections

    # Subtract all detections found in the user defined suspect file
    if suspect_file:
        print("Found suspect file {0}. Subtracting detections from input".format(suspect_file))
        susp_dets = pd.read_csv(suspect_file)
        good_dets = pd.concat([df, susp_dets], ignore_index=True)
        good_dets.drop_duplicates(mandatory_columns, keep=False, inplace=True)

    elif mandatory_columns.issubset(df.columns):
        # calculate detections to filtered

        # For each individual catalognumber:
        # Determine the space between each detection.
        # If the space before + after > min_time_buffer
        # Remove that detection row from the df and add it to suspect df.
        # SQL that does this is in load_to_postgresql under createSuspect

        ind = df['catalognumber'].unique()
        df['datecollected'] = pd.to_datetime(df['datecollected'])
        user_int = datetime.timedelta(minutes=min_time_buffer)
        good_dets = pd.DataFrame()
        susp_dets = pd.DataFrame()
        grouped = df.groupby('catalognumber')
        for anm in ind:
            anm_dets = grouped.get_group(anm)
            intervals = anm_dets['datecollected'] - anm_dets['datecollected'].shift(1)
            post_intervals = anm_dets['datecollected'].shift(-1) - anm_dets['datecollected']


            good_dets = good_dets.append(anm_dets[(intervals <= user_int) | (post_intervals <= user_int)])

        # If they aren't a good det, they're suspect!
        # TODO: Reporting: Decide if we want to report the big 'before/after' triplicate in Suspect Dets
        # If so, building susp_dets gets tougher, involves a merge and then a append.
        # For now, just a matter of putting the complement of the good dets in the susp_dets
        susp_dets = df[~df['unqdetecid'].isin(good_dets['unqdetecid'])]

    else:
        raise GenericException("Missing required input columns: {}".format(mandatory_columns - set(df.columns)))

    print("Total detections in filtered dataframe: {0}".format(len(good_dets.index)))
    print("{0} suspect detections removed".format(len(susp_dets.index)))

    output_dict = {"filtered": good_dets, "suspect": susp_dets}

    if distance_matrix:

        # Must now have lat and long columns as well.
        dm_mandatory_columns = set(['latitude', 'longitude'])
        if dm_mandatory_columns.issubset(df.columns):
            output_dict['dist_mtrx'] = get_distance_matrix(df)
            print("There are {0} station locations in the distance matrix".format(len(output_dict['dist_mtrx'].index)))
        else:
            raise GenericException("Missing required input columns for distance matrix calc: {}".format(dm_mandatory_columns - set(df.columns)))

    return output_dict
